/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jan 27, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.uvos.webapp.actions;

import java.io.IOException;
import java.net.MalformedURLException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import pl.edu.icm.pnpca.ws.CertificationAuthorityFault;
import pl.edu.icm.pnpca.ws.client.PnPCAClient;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.VOApplication;
import pl.edu.icm.unicore.uvos.api.VOApplicationForm;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.uvos.webapp.actions.appform.AppFormData;
import pl.edu.icm.uvos.webapp.actions.idform.IdFormData;
import pl.edu.icm.uvos.webapp.data.DataStore;
import xmlbeans.pl.edu.icm.uvos.internal.AttributeXType;
import xmlbeans.pl.edu.icm.uvos.internal.GroupXType;
import xmlbeans.pl.edu.icm.uvos.internal.app.CAAddressType;
import xmlbeans.pl.edu.icm.uvos.internal.app.VOApplicationFormConstraintsType;
import xmlbeans.pl.edu.icm.uvos.internal.app.VOApplicationRequestsDocument;
import xmlbeans.pl.edu.icm.uvos.internal.app.VOApplicationRequestsType;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.multipart.ByteArrayPartSource;
import org.apache.commons.httpclient.methods.multipart.FilePart;
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity;
import org.apache.commons.httpclient.methods.multipart.Part;
import org.apache.commons.httpclient.methods.multipart.StringPart;
import org.apache.commons.httpclient.protocol.Protocol;
import org.apache.commons.httpclient.protocol.ProtocolSocketFactory;
import org.apache.xmlbeans.XmlException;
import org.codehaus.xfire.XFireRuntimeException;
import org.codehaus.xfire.transport.http.EasySSLProtocolSocketFactory;
import org.springframework.webflow.action.AbstractAction;
import org.springframework.webflow.core.collection.MutableAttributeMap;
import org.springframework.webflow.execution.Event;
import org.springframework.webflow.execution.RequestContext;

/**
 * @author K. Benedyczak
 */
public class SubmitAction extends AbstractAction
{
	private DataStore dao;
	
	public SubmitAction(DataStore dao)
	{
		this.dao = dao;
	}
	
	@SuppressWarnings("unchecked")
	protected Event doExecute(RequestContext context) throws Exception
	{
		MutableAttributeMap scope = context.getFlowScope();
		String idS = (String) scope.get("formId");
		int id = Integer.parseInt(idS);
		
		Map<String, Object> appMap =  (Map<String, Object>) 
				scope.get("application");
		IdFormData idFormData =  (IdFormData) scope.get("app2FormData");
		AppFormData appFormData =  (AppFormData) scope.get("app1FormData");
		boolean isCSR = false;
		if (appFormData.getIsCSR() != null && appFormData.getIsCSR().size() > 0)
			isCSR = true;

		VOApplicationForm voForm = dao.getApplicationFormById(id);
		VOApplicationFormConstraintsType constraints;		
		try
		{
			constraints = VOApplicationFormConstraintsType.Factory.parse(
					voForm.getConstraints());
		} catch (XmlException e)
		{
			throw new InternalException("VO application form's " +
					"constraints are invalid");
		} 

		String contactEmail = appFormData.getEmail();
		if (contactEmail == null || contactEmail.equals(""))
			contactEmail = idFormData.getEmail();
		
		if (isCSR)
		{
			CAAddressType caaddr = constraints.getAllowCSR();
			try
			{
				if (caaddr.getNewPnPCA())
					processCSR4NewPnPCA(idFormData.getIdCert(), 
						caaddr.getAddress(), contactEmail);
				else
					processCSR(idFormData.getIdCert(), caaddr.getAddress());
			} catch (HTTPConnectionException e)
			{
				MutableAttributeMap flash = context.getFlashScope();
				flash.put("csrFailure", "Connection to CA failed. " +
						"The reason is: " + e.getContents());
				flash.put("caHost", caaddr.getAddress());
				return result("failure");
			} catch (CSRRejectedException e)
			{
				MutableAttributeMap flash = context.getFlashScope();
				flash.put("csrFailure", 
					"It seems that CA rejected your CSR. " +
					"The reason is: " + e.getContents());
				flash.put("caHost", caaddr.getAddress());
				return result("failure");
			} catch (ParseException e)
			{
				MutableAttributeMap flash = context.getFlashScope();
				flash.put("csrFailure", 
					"Failed to parse the response from the CA. " +
					"It may even happen that the request was accepted. " +
					"Parsing failed on: " + e.getMessage());
				flash.put("caHost", caaddr.getAddress());
				return result("failure");
			} catch (Exception e)
			{
				MutableAttributeMap flash = context.getFlashScope();
				flash.put("csrFailure", 
					"Failed to connect to the CA. " +
					"Details are: " + e);
				flash.put("caHost", caaddr.getAddress());
				return result("failure");	
			}
		}
		String format = (String) appMap.get("selectedFormat");
		Identity identity;
		if (format.equals(IdentityType.DN.getName()))
		{
			identity = new Identity(IdentityType.DN, 
					isCSR ? idFormData.getIdCert() :
						idFormData.getIdDN());
		} else if (format.equals(IdentityType.EMAIL.getName()))
		{
			identity = new Identity(IdentityType.EMAIL, 
					idFormData.getEmail());
			identity.setAuthnData(idFormData.getPasswd1());
		} else
		{
			identity = new Identity(IdentityType.X509CERT, 
					idFormData.getIdCert());
		}
		
		VOApplicationRequestsDocument xmlDoc = 
			VOApplicationRequestsDocument.Factory.newInstance();
		VOApplicationRequestsType xml = xmlDoc.addNewVOApplicationRequests();
		
		if (!constraints.getForceAttributeSelection())
		{
			List<String> extraAttrs = appFormData.getAdditionalAttrs();
			List<AttributeXType> attrsList = new ArrayList<AttributeXType>();
			if (extraAttrs != null)
			{
				for (String code: extraAttrs)
				{
					int i = Integer.parseInt(code);
					attrsList.add(constraints.getAttributesSelectionArray(i));
				}
				AttributeXType []attrsArray = attrsList.toArray(
						new AttributeXType[attrsList.size()]); 
				xml.setRequestedAttributesArray(attrsArray);
			}
		} else
		{
			if (constraints.getAttributesSelectionArray() != null &&
					constraints.getAttributesSelectionArray().length > 0)
				xml.setRequestedAttributesArray(
					constraints.getAttributesSelectionArray());
		}
		
		if (!constraints.getForceGroupSelection())
		{
			List<String> extraGroups = appFormData.getAdditionalGroups();
			List<GroupXType> groupsList = new ArrayList<GroupXType>();
			if (extraGroups != null)
			{
				for (String code: extraGroups)
				{
					int i = Integer.parseInt(code);
					groupsList.add(constraints.getSubgroupsSelectionArray(i));
				}
				GroupXType []groupsArray = groupsList.toArray(
						new GroupXType[groupsList.size()]);
				xml.setRequestedSubgroupsArray(groupsArray);
			}
		} else
		{
			if (constraints.getSubgroupsSelectionArray() != null &&
					constraints.getSubgroupsSelectionArray().length > 0)
				xml.setRequestedSubgroupsArray(constraints.getSubgroupsSelectionArray());			
		}
		

		if (idFormData.getCaCert() != null)
			xml.setCACertificate(idFormData.getCaCert());
		if (appFormData.getInstitution() != null)
			xml.setInstitution(appFormData.getInstitution());
		if (appFormData.getNotes() != null)
			xml.setNotes(appFormData.getNotes());
		if (appFormData.getPhone() != null)
			xml.setPhone(appFormData.getPhone());

		
		VOApplication application = new VOApplication(-1, id, 
				identity, isCSR, contactEmail, xml.xmlText(), null);
		dao.submitApplication(application);
		return success();
	}

	private void processCSR4NewPnPCA(String csr, String pnpCAUrl, String email) 
		throws CSRRejectedException, MalformedURLException
	{
		PnPCAClient pnpClient = new PnPCAClient(pnpCAUrl);
		try
		{
			try
			{
				pnpClient.getService().sumbmitCertificateRequest(csr.getBytes(), 
					"", "", email);
			} catch (XFireRuntimeException e)
			{
				throw new CSRRejectedException("Can't submit a CSR. Propably PnP CA version " +
						"is beta1 which is unsupported.");
			}
		} catch (CertificationAuthorityFault e)
		{
			throw new CSRRejectedException(e.getMessage());
		}
	}
	
	private void processCSR(String csr, String pnpCAUrl) 
		throws HttpException, IOException, ParseException, CSRRejectedException, HTTPConnectionException
	{
		Protocol.registerProtocol("myhttps", new Protocol("myhttps", 
			(ProtocolSocketFactory)new EasySSLProtocolSocketFactory(),
			8443));
		HttpClient client = new HttpClient();

		PostMethod filePost = new PostMethod(
				"myhttps://" + pnpCAUrl + "/requestsend.html");
		ByteArrayPartSource baps = new ByteArrayPartSource("csr.pem", 
				csr.getBytes());
		Part[] parts = 
		{
				new FilePart("request", baps),
				new StringPart("certtype", "unicore_user")
		};
		filePost.setRequestEntity(
				new MultipartRequestEntity(parts, filePost.getParams())
		);
		client.getHttpConnectionManager().
		getParams().setConnectionTimeout(7000);
		int status = client.executeMethod(filePost);
		if (status == HttpStatus.SC_OK) 
		{
			String resp = filePost.getResponseBodyAsString();
			String parsed = parse(resp);
			if (!isAccepted(parsed))
				throw new CSRRejectedException(parsed);
		} else 
		{
			throw new HTTPConnectionException(
					HttpStatus.getStatusText(status));
		}

	}


	private String parse(String html) throws ParseException
	{
		String startEl = "<div class=\"content\">";
		String endEl = "</div>";
		int i0 = html.indexOf(startEl);
		if (i0 == -1)
			throw new ParseException("No content element found", -1);
		html = html.substring(i0 + startEl.length());
		int iEnd = html.indexOf(endEl);
		if (iEnd == -1)
			throw new ParseException("Can't find the end of content " +
					"element", -1);
		html = html.substring(0, iEnd);
		return html;
	}

	private boolean isAccepted(String html)
	{
		return !html.contains("class=\"warning\"");
	}
	
	@SuppressWarnings("serial")
	private static class CSRRejectedException extends Exception
	{
		private String contents;
		
		public CSRRejectedException(String contents)
		{
			this.contents = contents;
		}

		public String getContents()
		{
			return contents;
		}
	}

	@SuppressWarnings("serial")
	private static class HTTPConnectionException extends Exception
	{
		private String contents;
		
		public HTTPConnectionException(String contents)
		{
			this.contents = contents;
		}

		public String getContents()
		{
			return contents;
		}
	}
}













